package com.tzld.piaoquan.ad.engine.service.score.strategy;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.TypeReference;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.ad.engine.commons.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.commons.enums.RedisPrefixEnum;
import com.tzld.piaoquan.ad.engine.commons.param.RankRecommendRequestParam;
import com.tzld.piaoquan.ad.engine.commons.redis.AdRedisHelper;
import com.tzld.piaoquan.ad.engine.commons.redis.AlgorithmRedisHelper;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.util.DateUtils;
import com.tzld.piaoquan.ad.engine.commons.util.NumUtil;
import com.tzld.piaoquan.ad.engine.commons.util.ObjUtil;
import com.tzld.piaoquan.ad.engine.service.entity.CalibrationCtcvrData;
import com.tzld.piaoquan.ad.engine.service.entity.CorrectCpaParam;
import com.tzld.piaoquan.ad.engine.service.entity.GuaranteeView;
import com.tzld.piaoquan.ad.engine.service.feature.Feature;
import com.tzld.piaoquan.ad.engine.service.feature.FeatureService;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

@Slf4j
public abstract class RankStrategyBasic implements RankStrategy {

    @ApolloJsonValue("${no.postback.conversion.adverids:[]}")
    private Set<String> noPostbackConversionAdVerIds;

    @ApolloJsonValue("${creative.model.score.coefficient:{}}")
    private Map<Long, Double> creativeScoreCoefficient;

    @Value("${guarantee.exp:742}")
    protected String guaranteeExp;
    @ApolloJsonValue("${alpha:1.0}")
    protected Double alpha;

    @Value("${calibration:ctcvr.exp:779}")
    protected String calibrationCtcvrExp;

    @ApolloJsonValue("${calibration.view.count:5000}")
    protected Integer calibrationViewCount;

    @Value("${correct.cpa.exp.1:}")
    protected String correctCpaExp1;

    @Value("${correct.cpa.exp.2:}")
    protected String correctCpaExp2;

    @Value("${correct.cpa.alpha.1:1.0}")
    protected Double correctCpaAlpha1;

    @Value("${correct.cpa.beta.1:1.0}")
    protected Double correctCpaBeta1;

    @Value("${correct.cpa.alpha.2:0.7}")
    protected Double correctCpaAlpha2;

    @Value("${correct.cpa.beta.2:0.2}")
    protected Double correctCpaBeta2;

    @Value("${correct.cpa.view:1000}")
    protected Integer correctCpaView;

    @Autowired
    private FeatureService featureService;
    @Autowired
    protected AdRedisHelper adRedisHelper;
    @Autowired
    protected AlgorithmRedisHelper algRedisHelper;

    String adPlatformGuaranteeKey = "ad:platform:guarantee:data:{date}:{adverId}";

    String realCtcvrCustomerKey = "ad:platform:real:ctcvr:{model}:{layer}:{class}:{profession}:{customerId}";

    String realCtcvrProfessionKey = "ad:platform:real:ctcvr:{model}:{layer}:{class}:{profession}";


    String adCustomerLayerHourKey = "ad:platform:customer:hour:{layer}:{clazz}:{customerId}";

    String adVerLayerHourKey = "ad:platform:adver:hour:{layer}:{clazz}:{adverId}";

    String adCustomerLayerDayKey = "ad:platform:customer:day:{layer}:{clazz}:{customerId}";

    String adVerLayerDayKey = "ad:platform:adver:day:{layer}:{clazz}:{adverId}";


    protected static final List<String> hasChannelScenes = new ArrayList<String>() {{
        add("DaiTou");
        add("GzhTouLiu");
        add("daitou");
        add("dyyjs");
        add("dyyqw");
        add("fwhdyy");
        add("gzhhz");
        add("gzhhzzx");
        add("shequn");
        add("touliu");
        add("xcxdt");
    }};


    protected Feature getFeature(ScoreParam param, RankRecommendRequestParam request) {
        List<AdPlatformCreativeDTO> adIdList = request.getAdIdList();
        List<String> cidList = adIdList.stream()
                .map(AdPlatformCreativeDTO::getCreativeId)
                .map(Object::toString)
                .collect(Collectors.toList());

        List<String> adVerIdList = adIdList.stream()
                .map(AdPlatformCreativeDTO::getAdVerId)
                .filter(StringUtils::isNotBlank)
                .distinct()
                .collect(Collectors.toList());

        List<Long> skuIdList = adIdList.stream()
                .map(AdPlatformCreativeDTO::getSkuId)
                .filter(Objects::nonNull)
                .distinct()
                .collect(Collectors.toList());
        return featureService.getFeature(cidList, adVerIdList, skuIdList, param);
    }

    protected Set<String> getNoApiAdVerIds() {
        Set<String> noApiAdVerIds = adRedisHelper.sMembers(RedisPrefixEnum.ADVER_IS_API_EQ_0_IDS.getPrefix());
        log.info("Redis中缓存的没有回传广告主ID集合: {}", noApiAdVerIds);
        if (CollectionUtils.isEmpty(noApiAdVerIds)) {
            noApiAdVerIds = noPostbackConversionAdVerIds;
        }
        return noApiAdVerIds;
    }

    protected Map<Long, Double> getCreativeScoreCoefficient() {
        if (Objects.isNull(creativeScoreCoefficient)) {
            return new HashMap<>();
        }
        return creativeScoreCoefficient;
    }

    protected Map<String, String> getReqFeature(ScoreParam scoreParam, RankRecommendRequestParam request) {
        Map<String, String> map = new HashMap<>();
        map.put("brand", scoreParam.getRequestContext().getMachineinfoBrand().equalsIgnoreCase("-1") ?
                "" : scoreParam.getRequestContext().getMachineinfoBrand().toUpperCase());
        map.put("region", Objects.equals(scoreParam.getRequestContext().getRegion(), "-1") ?
                "" : scoreParam.getRequestContext().getRegion());
        map.put("city", Objects.equals(scoreParam.getRequestContext().getCity(), "-1") ?
                "" : scoreParam.getRequestContext().getCity());
        map.put("vid", String.valueOf(request.getVideoId()));
        map.put("apptype", String.valueOf(request.getAppType()));
        String rootSessionId = request.getRootSessionId();
        String sessionId = request.getSessionId();
        String subSessionId = request.getSubSessionId();
        String rootSourceId = request.getRootSourceId();
        int isFirstLayer;
        if (Objects.equals(rootSessionId, subSessionId) || Objects.equals(rootSessionId, sessionId)) {
            isFirstLayer = 1;
        } else {
            isFirstLayer = 2;
        }
        map.put("is_first_layer", String.valueOf(isFirstLayer));
        if (StringUtils.isNotEmpty(rootSourceId)) {
            String rootSourceScene = rootSourceId.split("_")[0];
            map.put("root_source_scene", rootSourceScene);
            if (hasChannelScenes.contains(rootSourceScene) && rootSourceId.split("_").length > 1) {
                String rootSourceChannel = rootSourceId.split("_")[1];
                map.put("root_source_channel", rootSourceChannel);
            }
        }
        return map;
    }

    protected Map<String, String> handleSceneFeature(long ts) {
        Map<String, String> sceneFeatureMap = new HashMap<>();
        sceneFeatureMap.put("hour_" + DateUtils.getHourByTimestamp(ts), "0.1");
        sceneFeatureMap.put("dayofweek_" + DateUtils.getDayOrWeekByTimestamp(ts), "0.1");
        sceneFeatureMap.put("hour", String.valueOf(DateUtils.getHourByTimestamp(ts)));
        sceneFeatureMap.put("hour_quarter", String.valueOf(DateUtils.getHourQuarter(ts)));
        sceneFeatureMap.put("ts", String.valueOf(ts));
        return sceneFeatureMap;
    }


    protected static class Tuple5 {
        public String f1;
        public String f2;
        public String f3;
        public String f4;
        public String f5;

        public Tuple5(String f1, String f2, String f3, String f4, String f5) {
            this.f1 = f1;
            this.f2 = f2;
            this.f3 = f3;
            this.f4 = f4;
            this.f5 = f5;
        }
    }

    protected static class TupleMapEntry<T> {
        public String key;
        public T value;

        public TupleMapEntry(String key, T value) {
            this.key = key;
            this.value = value;
        }
    }

    protected static class Tuple2<F1, F2> {
        public F1 f1;

        public F2 f2;

        public Tuple2(F1 first, F2 name) {
            this.f1 = first;
            this.f2 = name;
        }

    }

    protected Map<String, GuaranteeView> getGuaranteeViewMap(RankRecommendRequestParam request, ScoreParam scoreParam) {
        Map<String, GuaranteeView> map = new HashMap<>();
        try {
            if (scoreParam.getExpCodeSet().contains(guaranteeExp)) {
                String thatDayDateString = DateUtils.getThatDayDateString();
                String redisKey = adPlatformGuaranteeKey.replace("{date}", thatDayDateString);
                List<String> adVerIds = request.getAdIdList().stream().map(AdPlatformCreativeDTO::getAdVerId).distinct()
                        .filter(Objects::nonNull).collect(Collectors.toList());
                List<String> redisKeys = adVerIds.stream().map(e -> redisKey.replace("{adverId}", e)).collect(Collectors.toList());
                List<String> values = adRedisHelper.mget(redisKeys);
                if (CollectionUtils.isNotEmpty(values)) {
                    for (int i = 0; i < redisKeys.size(); i++) {
                        String value = values.get(i);
                        if (value != null) {
                            GuaranteeView guaranteeView = JSONObject.parseObject(value, GuaranteeView.class);
                            map.put(adVerIds.get(i), guaranteeView);
                        }
                    }
                }
            }
        } catch (Exception e) {
            log.error("getGuaranteeViewMap error", e);
        }

        return map;
    }

    protected void setGuaranteeWeight(Map<String, GuaranteeView> map, String adVerId, Map<String, Object> ext) {
        if (MapUtils.isNotEmpty(map)) {
            GuaranteeView guaranteeView = map.get(adVerId);
            if (guaranteeView != null) {
                double guaranteeWeight = calculateGuaranteedWeight(guaranteeView);
                boolean isGuaranteed = isGuaranteed(guaranteeView);
                ext.put("guaranteeView", guaranteeView.toString());
                ext.put("guaranteeWeight", guaranteeWeight);
                ext.put("isGuaranteed", isGuaranteed);
            }
        }
    }

    protected double calculateGuaranteedWeight(GuaranteeView guaranteeView) {
        if (guaranteeView == null) {
            return 1.0;
        }
        double guaranteeWeight;
        if (guaranteeView.getGuaranteeNum() == null || guaranteeView.getGuaranteeNum() == 0
                || guaranteeView.getGuaranteeRate() == null || guaranteeView.getGuaranteeRate() == 0.0) {
            guaranteeWeight = 1.0;
        } else {
            if (guaranteeView.getAdrAlgoViewNum() != null &&
                    guaranteeView.getGuaranteeNum() <= guaranteeView.getAdrAlgoViewNum()) {
                guaranteeWeight = 0.0;
            } else {
                int allViewNum = guaranteeView.getAllAlgoViewNum() == null || guaranteeView.getAllAlgoViewNum() == 0 ?
                        1 : guaranteeView.getAllAlgoViewNum();
                int adrAlogViewNum = guaranteeView.getAdrAlgoViewNum() == null || guaranteeView.getAdrAlgoViewNum() == 0 ?
                        1 : guaranteeView.getAdrAlgoViewNum();
                // guaranteeView.getGuaranteeRate() 是百分之几  要乘0.01
                guaranteeWeight = guaranteeView.getGuaranteeRate() * 0.01 * allViewNum / adrAlogViewNum;
                if (guaranteeWeight < 0.5) {
                    guaranteeWeight = 0.5;
                }
                if (guaranteeWeight > 2) {
                    guaranteeWeight = 2;
                }
            }
        }
        return guaranteeWeight;
    }

    protected boolean isGuaranteed(GuaranteeView guaranteeView) {
        if (guaranteeView == null) {
            return false;
        }
        return guaranteeView.getGuaranteeNum() != null && guaranteeView.getGuaranteeNum() != 0
                && guaranteeView.getGuaranteeRate() != null && guaranteeView.getGuaranteeRate() != 0.0;
    }

    protected double getGuaranteeScoreCoefficient(ScoreParam scoreParam, Map<String, Object> ext) {
        if (scoreParam.getExpCodeSet().contains(guaranteeExp)) {
            if (ext.get("guaranteeWeight") == null) {
                return 1.0;
            } else {
                return Math.pow((double) ext.get("guaranteeWeight"), alpha);
            }
        } else {
            return 1.0;
        }
    }

    protected Map<Long, CorrectCpaParam> getCorrectCpaParamMap(RankRecommendRequestParam request, ScoreParam scoreParam) {
        Map<String, String> userLayer = this.getUserLayer(request.getMid());

        String layer = userLayer.getOrDefault("layer", "无曝光");
        String clazz = userLayer.getOrDefault("class", "近期未出现");
        if (StringUtils.isNotEmpty(layer) && layer.equals("已转化")) {
            layer = "有转化";
        }
        if (request.getIsFilterUser()) {
            layer = layer + "-炸";
        }

        String redisAdCustomerLayerHourKey = adCustomerLayerHourKey.replace("{layer}", layer).replace("{clazz}", clazz);
        String redisAdVerLayerHourKey = adVerLayerHourKey.replace("{layer}", layer).replace("{clazz}", clazz);
        String redisAadCustomerLayerDayKey = adCustomerLayerDayKey.replace("{layer}", layer).replace("{clazz}", clazz);
        String redisAdVerLayerDayKey = adVerLayerDayKey.replace("{layer}", layer).replace("{clazz}", clazz);


        Map<Long, CorrectCpaParam> resultMap = new HashMap<>();
        try {
            if (CollectionUtils.isEmpty(request.getAdIdList())) {
                return resultMap;
            }

            // 抽取公共方法获取数据
            List<Long> customerIds = request.getAdIdList().stream()
                    .map(AdPlatformCreativeDTO::getCustomerId)
                    .distinct()
                    .collect(Collectors.toList());
            Map<Long, JSONObject> customerLayerMap = getRedisData(
                    customerIds,
                    id -> redisAdCustomerLayerHourKey.replace("{customerId}", String.valueOf(id)),
                    id -> redisAadCustomerLayerDayKey.replace("{customerId}", String.valueOf(id))
            );

            List<String> adVerIds = request.getAdIdList().stream()
                    .map(AdPlatformCreativeDTO::getAdVerId)
                    .distinct()
                    .collect(Collectors.toList());
            Map<String, JSONObject> adVerLayerMap = getRedisData(
                    adVerIds,
                    id -> redisAdVerLayerHourKey.replace("{adverId}", id),
                    id -> redisAdVerLayerDayKey.replace("{adverId}", id)
            );

            for (AdPlatformCreativeDTO adPlatformCreativeDTO : request.getAdIdList()) {
                Long creativeId = adPlatformCreativeDTO.getCreativeId();
                Long customerId = adPlatformCreativeDTO.getCustomerId();
                JSONObject jsonObject;
                if (customerId != null && customerId != 0 && customerLayerMap.containsKey(customerId)) {
                    jsonObject = customerLayerMap.get(customerId);
                } else {
                    jsonObject = adVerLayerMap.get(adPlatformCreativeDTO.getAdVerId());
                }
                CorrectCpaParam correctCpaParam = new CorrectCpaParam();
                if (jsonObject == null) {
                    correctCpaParam.setCorrectionFactor(1.0);
                    resultMap.put(creativeId, correctCpaParam);
                    continue;
                }
                Integer views = jsonObject.getInteger("viewsHour");
                Double realCtcvrHour = jsonObject.getDouble("realCtcvrHour");
                Double pCtcvrHour = jsonObject.getDouble("pCtcvrHour");
                Double realCtcvrDay = jsonObject.getDouble("realCtcvrDay");
                Double pCtcvrDay = jsonObject.getDouble("pCtcvrDay");
                correctCpaParam.setRealCtcvrHour(realCtcvrHour);
                correctCpaParam.setPCtcvrHour(pCtcvrHour);
                correctCpaParam.setRealCtcvrDay(realCtcvrDay);
                correctCpaParam.setPCtcvrDay(pCtcvrDay);
                correctCpaParam.setView(views);
                double correctionFactor = 1.0;
                //曝光数小于目标曝光数，不进行修正
                if (views == null || views < correctCpaView) {
                    correctCpaParam.setCorrectionFactor(correctionFactor);
                    resultMap.put(creativeId, correctCpaParam);
                    continue;
                }
                if (realCtcvrHour != null && pCtcvrHour != null && realCtcvrDay != null && pCtcvrDay != null) {
                    if (scoreParam.getExpCodeSet().contains(correctCpaExp2)) {
                        correctionFactor = (1 - correctCpaAlpha2 - correctCpaBeta2) +
                                NumUtil.div(realCtcvrHour, pCtcvrHour) * correctCpaAlpha2 +
                                NumUtil.div(realCtcvrDay, pCtcvrDay) * correctCpaBeta2;
                    } else {
                        correctionFactor = Math.pow(NumUtil.div(realCtcvrHour, pCtcvrHour), correctCpaAlpha1) *
                                Math.pow(NumUtil.div(realCtcvrDay, pCtcvrDay), correctCpaBeta1);
                    }
                    if (correctionFactor <= 0) {
                        correctionFactor = 1;
                    }
                }
                correctCpaParam.setCorrectionFactor(correctionFactor);
                resultMap.put(creativeId, correctCpaParam);
            }
        } catch (Exception e) {
            log.error("getCorrectCpaParamMap error", e);
        }
        return resultMap;
    }

    private <T> Map<T, JSONObject> getRedisData(
            Collection<T> ids,
            Function<T, String> hourKeyBuilder,
            Function<T, String> dayKeyBuilder) {

        Map<T, JSONObject> resultMap = new HashMap<>();
        if (CollectionUtils.isEmpty(ids)) {
            return resultMap;
        }

        // 构建Keys
        List<String> hourKeys = ids.stream()
                .map(hourKeyBuilder)
                .collect(Collectors.toList());
        List<String> dayKeys = ids.stream()
                .map(dayKeyBuilder)
                .collect(Collectors.toList());

        // 批量获取Redis值
        List<String> hourValues = adRedisHelper.mget(hourKeys);
        List<String> dayValues = adRedisHelper.mget(dayKeys);

        // 解析数据
        Iterator<T> idIter = ids.iterator();
        for (int i = 0; i < ids.size(); i++) {
            T id = idIter.next();
            JSONObject jsonObj = new JSONObject();

            // 解析小时数据
            parseRedisValue(hourValues.get(i), jsonObj, "Hour", "ctcvr", "pCtcvr", "views");
            // 解析天数据
            parseRedisValue(dayValues.get(i), jsonObj, "Day", "ctcvr", "pCtcvr", "views");

            resultMap.put(id, jsonObj);
        }
        return resultMap;
    }

    private void parseRedisValue(String value, JSONObject target,
                                 String suffix, String... keys) {
        if (StringUtils.isEmpty(value)) return;

        try {
            JSONObject source = JSONObject.parseObject(value);
            for (String key : keys) {
                Object val = source.get(key);
                if (val != null) {
                    String newKey = (key.equals("ctcvr") ? "realCtcvr" : key) + suffix;
                    target.put(newKey, val);
                }
            }
        } catch (Exception e) {
            log.warn("Parse redis value failed: {}", value, e);
        }
    }


    protected AdRankItem creativeCovertRankItem(AdPlatformCreativeDTO dto, RankRecommendRequestParam request, Set<String> noApiAdVerIds) {
        AdRankItem adRankItem = new AdRankItem();
        adRankItem.setAdId(dto.getCreativeId());
        adRankItem.setCreativeCode(dto.getCreativeCode());
        adRankItem.setAdVerId(dto.getAdVerId());
        adRankItem.setVideoId(request.getVideoId());
        adRankItem.setCpa(dto.getCpa());
        adRankItem.setId(dto.getAdId());
        adRankItem.setCampaignId(dto.getCampaignId());
        adRankItem.setCpm(ObjUtil.nullOrDefault(dto.getCpm(), 90).doubleValue());
        adRankItem.setSkuId(dto.getSkuId());
        adRankItem.getExt().put("recallsources", dto.getRecallSources());
        adRankItem.setRandom(new Random().nextInt(1000));
        if (CollectionUtils.isNotEmpty(noApiAdVerIds)) {
            if (noApiAdVerIds.contains(dto.getAdVerId())) {
                adRankItem.getExt().put("isApi", "0");
            } else {
                adRankItem.getExt().put("isApi", "1");
            }
        }

        return adRankItem;
    }

    protected void putMetaFeature(AdRankItem adRankItem, Feature feature, Map<String, String> reqFeature,
                                  Map<String, String> sceneFeatureMap, RankRecommendRequestParam request) {
        Map<String, Map<String, String>> userFeature = feature.getUserFeature();
        Map<String, Map<String, String>> videoFeature = feature.getVideoFeature();
        Map<String, Map<String, Map<String, String>>> allAdVerFeature = feature.getAdVerFeature();
        Map<String, Map<String, Map<String, String>>> allCidFeature = feature.getCidFeature();
        Map<String, Map<String, Map<String, String>>> allSkuFeature = feature.getSkuFeature();
        for (Map.Entry<String, Map<String, String>> entry : videoFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        for (Map.Entry<String, Map<String, String>> entry : userFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        Map<String, Map<String, String>> adVerFeature = allAdVerFeature.getOrDefault(adRankItem.getAdVerId(), new HashMap<>());
        for (Map.Entry<String, Map<String, String>> entry : adVerFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        Map<String, Map<String, String>> cidFeature = allCidFeature.getOrDefault(String.valueOf(adRankItem.getAdId()), new HashMap<>());
        for (Map.Entry<String, Map<String, String>> entry : cidFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }

        Map<String, Map<String, String>> skuFeature = allSkuFeature.getOrDefault(String.valueOf(adRankItem.getSkuId()), new HashMap<>());
        for (Map.Entry<String, Map<String, String>> entry : skuFeature.entrySet()) {
            if (MapUtils.isNotEmpty(entry.getValue())) {
                adRankItem.getMetaFeatureMap().put(entry.getKey(), entry.getValue());
            }
        }
        long cid = adRankItem.getAdId();
        List<AdPlatformCreativeDTO> adIdList = request.getAdIdList();
        AdPlatformCreativeDTO adPlatformCreativeDTO = adIdList.stream()
                .filter(dto -> cid == dto.getCreativeId())
                .findFirst().orElse(null);
        if (adPlatformCreativeDTO != null) {
            reqFeature.put("cid", String.valueOf(adPlatformCreativeDTO.getCreativeId()));
            reqFeature.put("adid", String.valueOf(adPlatformCreativeDTO.getAdId()));
            reqFeature.put("adverid", String.valueOf(adPlatformCreativeDTO.getAdVerId()));
            reqFeature.put("profession", adPlatformCreativeDTO.getProfession());
        }
        adRankItem.getMetaFeatureMap().put("reqFeature", reqFeature);
        adRankItem.getMetaFeatureMap().put("sceneFeature", sceneFeatureMap);
    }

    protected Map<String, String> getUserLayer(String mid) {
        if (StringUtils.isEmpty(mid)) {
            return new HashMap<>();
        }
        String key = String.format("ad:engine:mid:layer:%s", mid);
        String value = algRedisHelper.get(key);
        if (StringUtils.isEmpty(value)) {
            return new HashMap<>();
        }
        try {
            return JSON.parseObject(value, new TypeReference<Map<String, String>>() {
            });
        } catch (Exception e) {
            log.error("getUserLayer error. mid: {}, \n", mid, e);
            return new HashMap<>();
        }
    }

    protected void calibrationCtcvrScore(ScoreParam scoreParam, List<AdRankItem> adRankItems, String mid, boolean isFilterUser, String modelName) {
        //判断是否走校准试验
        if (!scoreParam.getExpCodeSet().contains(calibrationCtcvrExp)) {
            return;
        }
        // 1. 获取用户分层信息
        Map<String, String> userLayer = getUserLayer(mid);
        String layer = userLayer.getOrDefault("layer", "无曝光");
        String clazz = userLayer.getOrDefault("class", "近期未出现");
        if (isFilterUser) {
            layer += "-炸";
        }

        // 2. 构建Key模板
        String customerKeyTemplate = realCtcvrCustomerKey
                .replace("{model}", modelName)
                .replace("{layer}", layer)
                .replace("{class}", clazz);

        String professionKeyTemplate = realCtcvrProfessionKey
                .replace("{model}", modelName)
                .replace("{layer}", layer)
                .replace("{class}", clazz);

        // 3. 定义Key生成器
        Function<AdRankItem, String> customerKeyFunc = e ->
                (e.getProfession() != null && e.getCustomerId() != null)
                        ? customerKeyTemplate.replace("{profession}", e.getProfession())
                        .replace("{customerId}", String.valueOf(e.getCustomerId()))
                        : null;

        Function<AdRankItem, String> professionKeyFunc = e ->
                (e.getProfession() != null)
                        ? professionKeyTemplate.replace("{profession}", e.getProfession())
                        : null;

        // 4. 获取校准数据
        Map<String, CalibrationCtcvrData> customerMap = getCtcvrMap(adRankItems, customerKeyFunc);
        Map<String, CalibrationCtcvrData> professionMap = getCtcvrMap(adRankItems, professionKeyFunc);
        log.info("calibrationCtcvrScore customerMap={}", JSON.toJSONString(customerMap));
        log.info("calibrationCtcvrScore professionMap={}", JSON.toJSONString(professionMap));
        // 5. 校准分数
        for (AdRankItem item : adRankItems) {
            CalibrationCtcvrData data = null;
            String customerKey = customerKeyFunc.apply(item);
            String professionKey = professionKeyFunc.apply(item);

            if (customerKey != null) {
                data = customerMap.get(customerKey);
            }
            if (data == null && professionKey != null) {
                data = professionMap.get(professionKey);
            }
            item.getExt().put("calibrationCtcvrData", JSONObject.toJSONString(data));
            if (data == null || data.getView() == null || data.getView() < calibrationViewCount) continue;

            Map<String, Double> scoreMap = item.getScoreMap();
            Double pCtcvr = data.getPCtcvr();
            Double realCtcvr = data.getRealCtcvr();

            if (pCtcvr != null && pCtcvr != 0.0 && realCtcvr != null && realCtcvr != 0.0) {
                double diff = realCtcvr / pCtcvr;
                if (Math.abs(diff - 1) >= 0.1) {
                    Double ctcvrScore = scoreMap.get("ctcvrScore");
                    if (ctcvrScore == null) {
                        continue;
                    }
                    scoreMap.put("modelCtcvrScore", ctcvrScore);
                    scoreMap.put("ctcvrScore", ctcvrScore * diff);
                    item.setLrScore(ctcvrScore * diff);
                }
            }
        }
    }

    // 通用CTR数据获取方法
    private Map<String, CalibrationCtcvrData> getCtcvrMap(
            List<AdRankItem> items,
            Function<AdRankItem, String> keyGenerator
    ) {
        // 生成有效Key集合
        List<String> redisKeys = items.stream()
                .map(keyGenerator)
                .filter(Objects::nonNull)
                .distinct()
                .collect(Collectors.toList());

        // 批量查询Redis
        List<String> redisValues = adRedisHelper.mget(redisKeys);
        Map<String, CalibrationCtcvrData> resultMap = new HashMap<>();

        for (int i = 0; i < redisKeys.size(); i++) {
            String val = redisValues.get(i);
            if (StringUtils.isEmpty(val)) continue;

            try {
                JSONObject json = JSONObject.parseObject(val);
                Integer view = json.getInteger("view");
                Double ctcvr = json.getDouble("ctcvr");
                Double pCtcvr = json.getDouble("pCtcvr");
                CalibrationCtcvrData data = new CalibrationCtcvrData();
                data.setView(view);
                data.setRealCtcvr(ctcvr);
                data.setPCtcvr(pCtcvr);
                resultMap.put(redisKeys.get(i), data);
            } catch (Exception e) {
                log.error("getCtcvrMap error", e);
            }
        }
        return resultMap;
    }
}
